#!/usr/bin/perl
use strict;

# usage:  mrinfo-validate.pl sets-to-validate mrinfo-data...

# config
my $sort_response = 1;

# state
my @alias_sets = ();
my $dst = undef;
my $meta = undef;
my @lines = ();
my %response_ifaces = ();
my %responses;
my %ifaces;
my %dsts;
my $n_empty_resp = 0;
my @similarities = ();

my $n_pairs = 0;
my $n_no_mrinfo = 0;
my $n_sim_0 = 0;
my $n_sim_1 = 0;
my $n_sim_inbetween = 0;
my $current_set;

sub isBogon($)
{
    my ($addr) = @_;
    return
	$addr =~ /0\./ ||                          # this network
	$addr =~ /127\./ ||                        # loopback
	$addr =~ /169\.254\./ ||                   # link local (RFC3330)
	$addr =~ /192\.0\.2\./ ||                  # test net (RFC3330)
	$addr =~ /198\.1[89]\./ ||                 # benchmark (RFC3330)
	$addr =~ /(22[4-9]|23\d)\./ ||             # 224/8 - 239/8 multicast
	$addr =~ /255\./ ||                        # broadcast
	$addr =~ /10\./ ||                         # private - RFC 1918
	$addr =~ /172\.(1[6-9]|2\d|3[01]|)\./ ||   # private - RFC 1918
	$addr =~ /192\.168\./;                     # private - RFC 1918
}

sub response_similarity($$)
{
    my ($r1, $r2) = @_;
    return 1 if ($r1 eq $r2);
    my $ref1 = $responses{$r1};
    my $ref2 = $responses{$r2};
    my %r1ifaces = ();
    my $shared = 0;
    my $total = 0;
    # print "## r1 ifaces: ", $ref1->{ifaces} ? join(', ', @{$ref1->{ifaces}}) : "undef", "\n";
    # print "## r2 ifaces: ", $ref2->{ifaces} ? join(', ', @{$ref2->{ifaces}}) : "undef", "\n";
    for my $iface (@{$ref1->{ifaces}}) {
	$r1ifaces{$iface} = 1;
	$total++
    }
    for my $iface (@{$ref2->{ifaces}}) {
	if ($r1ifaces{$iface}) {
	    $shared++;
	} else {
	    $total++;
	}
    }
    # print "## similarity = $shared/$total\n";
    return ($total == 0 ? 0 : $shared/$total) . " ($shared/$total)";
}

sub endresponse()
{
    if (scalar keys %response_ifaces == 0) {
	$n_empty_resp++;
    } else {
	my $response = $sort_response ? join('', sort @lines) : join('', @lines);
	$responses{$response}{count}++;
	my $resp_obj = $responses{$response};
	if ($resp_obj->{count} == 1) {
	    $resp_obj->{meta} = $meta;
	    @{$resp_obj->{ifaces}} = sort keys %response_ifaces;
	    # print "#### ifaces: ", join(', ', @{$resp_obj->{ifaces}}), "\n";
	    for my $iface (keys %response_ifaces) {
		next if (isBogon($iface));
		push @{$ifaces{$iface}}, $response;
	    }
	}
	$dsts{$dst}{n_raw_response}++;
	if ($dsts{$dst}{response}[0] ne $response) {
	    $dsts{$dst}{n_response}++;
	    my $n = $dsts{$dst}{n_response};
	    if ($n > 1) {
		my $sim = response_similarity($dsts{$dst}{response}[0], $response);
		my $metasim =
		    $responses{$dsts{$dst}{response}[0]}{meta} eq $responses{$response}{meta} ?
		    "same" : "different";
		push @similarities, 0+$sim;
		# print "## $dst similarity of response 1 vs $n: $sim; meta: $metasim\n";
		if ($dsts{$dst}{n_response} == 2) {
		    # print "## $dst response #1:\n";
		    # print $dsts{$dst}{response}[0];
		}
		# print "## $dst response #$n:\n";
		# print $response;
		# print "##\n";
	    }
	    push @{$dsts{$dst}{response}}, $response;
	}
	push @{$responses{$response}{dsts}}, $dst;
	#print "########\n";
	#print "response:\n$response";
	#print "dst: $dst\n";
	#print "ifaces: ", join(', ', keys %response_ifaces), "\n";
    }

    @lines = ();
    $dst = undef;
    $meta = undef;
    %response_ifaces = ();
}

sub histogram($$)
{
    my ($dataref, $label) = @_;
    print "# histogram of $label\n";
    for (my $i = 0; $i < scalar @$dataref; $i++) {
	next unless ($dataref->[$i]);
	printf "%4d %6d\n", $i, $dataref->[$i];
    }
}

open(SETS, "<", $ARGV[0]) or die "$ARGV[0]: $!\n";
while (<SETS>) {
    if (/^# set /) {
	$current_set = [];
	push @alias_sets, $current_set;
    } elsif (/^(\d+\.\d+\.\d+\.\d+)$/) {
	push @$current_set, $1;
    }
}
shift @ARGV;

while (<>) {
    #print "######## line: $_";
    if (/^\S/) {
	endresponse if (defined $dst);
    }

#    if (/^DVMRP response from (\d+\.\d+\.\d+\.\d+):/) {
#	$dst = $1;
#	$dsts{$dst}{n_error}++;
#	push @{$dsts{$dst}{response}}, $_;
#	$dst = undef;
#    } elsif (/^fail (\d+\.\d+\.\d+\.\d+)/) {
#	$dst = $1;
#	$dsts{$dst}{n_fail}++;
#	push @{$dsts{$dst}{response}}, $_;
#	$dst = undef;

    # Handle two slightly different formats:
    #  1. tcpdump: 193.137.55.201 [version 12.2]:
    #  2. mrinfo:  124.214.85.193 (NaN) [version 12.2,prune,mtrace]:
    if (/^(\d+\.\d+\.\d+\.\d+)(?: \(NaN\))?/) {
	#print "# new response\n";
	$dst = $1;
	@lines = ($');
        $meta = $';
    } elsif (/^ +0\.0\.0\.0 ->/) {
	# ignore
#    } elsif (/^ +(\d+\.\d+\.\d+\.\d+) -> 0\.0\.0\.0/) {
#	# ignore
    } elsif (/^ +(\d+\.\d+\.\d+\.\d+) ->/) {
	#print "# iface line\n";
        my $addr = $1;
	my $count = 1;

        # The combine-mrinfo-responses script collapses repeated lines into
        # a single line ending with a count.  For example,
        #     0.0.0.0 -> 0.0.0.0 (1/0/disabled) 3
	if (/ (\d+)$/) {
	  $count = $1;
	  s/ \d+$//;
	}

	$response_ifaces{$addr} += $count;
	while ($count > 0) {
	  push @lines, $_;
	  $count -= 1;
	}
    }
}
endresponse() if (defined $dst);

sub compare($$)
{
    my ($if1, $if2) = @_;
    my ($resp1, $resp2);
    $resp1 = $dsts{$if1};
    $resp2 = $dsts{$if2};
    if (!defined $resp1 || !defined $resp2) {
	$n_no_mrinfo++;
	return 0;
    }
    my $string1 = $resp1->{response}[0];
    my $string2 = $resp2->{response}[0];
    my $sim = response_similarity($string1, $string2);
    if ($sim == 0) {
	$n_sim_0++;
        print "zero similarity: $if1 and $if2\n", $string1, $string2;
    } elsif ($sim == 1) {
	$n_sim_1++;
    } else {
	$n_sim_inbetween++;
        print "partial similarity: $if1 and $if2, similarity = $sim\n", $string1, $string2;
    }
}

for my $alias_set (@alias_sets) {
    for (my $i = 0; $i < scalar @$alias_set; $i++) {
	for (my $j = $i+1; $j < scalar @$alias_set; $j++) {
	    $n_pairs++;
	    compare $alias_set->[$i], $alias_set->[$j];
	}
    }
}

printf "# number of alias sets in data set:           %7d\n", scalar @alias_sets;
printf "# number of alias pairs in data set:          %7d\n", $n_pairs;
printf "# number of pairs without mrinfo data:        %7d\n", $n_no_mrinfo;
printf "# number of pairs with mrinfo similarity=1:   %7d\n", $n_sim_1;
printf "# number of pairs with mrinfo similarity=0:   %7d\n", $n_sim_0;
printf "# number of pairs with mrinfo 0<similarity<1: %7d\n", $n_sim_inbetween;

